/*$Id: d_subckt.cc,v 25.39 2005/08/08 02:17:46 al Exp $ -*- C++ -*-
 * Copyright (C) 2001 Albert Davis
 * Author: Albert Davis <aldavis@ieee.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *------------------------------------------------------------------
 * subcircuit stuff
 * base class for other elements using internal subckts
 * netlist syntax:
 * device: Xxxxx <nodelist> <subckt-name>
 * model:  .subckt <subckt-name> <nodelist>
 *	   (device cards)
 *	   .ends <subckt-name>
 * storage note ...
 * the .subckt always has a comment at the hook point, so a for loop works
 * the expansion (attact to the X) has all comments removed
 *	- need to process the entire ring - for doesn't work
 */
#include "l_stlextra.h"
#include "c_comand.h" // ".ends"
#include "ap.h"
#include "e_model.h"
#include "d_subckt.h"
/*--------------------------------------------------------------------------*/
//		MODEL_SUBCKT::MODEL_SUBCKT(const char*);
//	void	MODEL_SUBCKT::parse_spice(CS& cmd);
// 	void	MODEL_SUBCKT::print(int,int)const;

//		DEV_SUBCKT::DEV_SUBCKT();
//		DEV_SUBCKT::DEV_SUBCKT(const DEV_SUBCKT& p);
//	void	DEV_SUBCKT::parse_spice(CS& cmd);
// 	void	DEV_SUBCKT::print(int,int)const;
//	void	DEV_SUBCKT::expand();
//	double	DEV_SUBCKT::tr_probe_num(CS&)const;
/*--------------------------------------------------------------------------*/
int DEV_SUBCKT::_count = 0;
int COMMON_SUBCKT::_count = -1;
int MODEL_SUBCKT::_count = 0;
static COMMON_SUBCKT Default_SUBCKT(CC_STATIC);
/*--------------------------------------------------------------------------*/
bool COMMON_SUBCKT::operator==(const COMMON_COMPONENT& x)const
{
  const COMMON_SUBCKT* p = dynamic_cast<const COMMON_SUBCKT*>(&x);
  return p && is_equal(x);
}
/*--------------------------------------------------------------------------*/
MODEL_SUBCKT::MODEL_SUBCKT()
  :COMPONENT(),
   _net_nodes(0)
{
  _n = _nodes;
  new_subckt();
  ++_count;
}
/*--------------------------------------------------------------------------*/
MODEL_SUBCKT::~MODEL_SUBCKT()
{
  --_count;
}
/*--------------------------------------------------------------------------*/
void MODEL_SUBCKT::parse_spice(CS& cmd)
{ 
  // parse the line
  cmd.reset();
  cmd.skiparg();	/* skip known ".subckt" */
  parse_label(cmd);
  _net_nodes = parse_nodes(cmd, max_nodes(), min_nodes(), 0 /*no tail*/);
}
/*--------------------------------------------------------------------------*/
void MODEL_SUBCKT::print_spice(OMSTREAM& where, int)const
{
  where << ".subckt " << short_label() << ' ';
  printnodes(where);
  where << '\n';

  {if (subckt()) {
    for (CARD_LIST::const_iterator 
	   ci = subckt()->begin(); ci != subckt()->end(); ++ci) {
      (**ci).print_spice(where, false);
    }
  }else{
    untested(); 
  }}
  where << ".ends " << short_label() << '\n';
}
/*--------------------------------------------------------------------------*/
DEV_SUBCKT::DEV_SUBCKT()
  :BASE_SUBCKT(),
   _net_nodes(0)
{
  attach_common(&Default_SUBCKT);
  _n = _nodes;
  ++_count;
}
/*--------------------------------------------------------------------------*/
DEV_SUBCKT::DEV_SUBCKT(const DEV_SUBCKT& p)
  :BASE_SUBCKT(p),
   _net_nodes(p._net_nodes)
{
  //strcpy(modelname, p.modelname); in common
  for (int ii = 0;  ii < max_nodes();  ++ii) {
    _nodes[ii] = p._nodes[ii];
  }
  _n = _nodes;
  ++_count;
}
/*--------------------------------------------------------------------------*/
void DEV_SUBCKT::parse_spice(CS& cmd)
{
  assert(has_common());
  COMMON_SUBCKT* c = prechecked_cast<COMMON_SUBCKT*>(common()->clone());
  assert(c);
  assert(!c->has_model());

  parse_Label(cmd);
  _net_nodes = parse_nodes(cmd, max_nodes(), min_nodes());
  c->parse_modelname(cmd);
  attach_common(c);
}
/*--------------------------------------------------------------------------*/
void DEV_SUBCKT::print_spice(OMSTREAM& where, int)const
{
  const COMMON_SUBCKT* c = prechecked_cast<const COMMON_SUBCKT*>(common());
  assert(c);
  where << short_label();
  printnodes(where);
  where << "  " << c->modelname() << '\n';
}
/*--------------------------------------------------------------------------*/
inline const CARD* find_subckt_model(CARD* scope, const std::string& modelname)
{
  const CARD* model = 0;
  {
    model = scope->find_looking_out(modelname, bDEBUG);
    if (!model) {
      error(bDANGER, "can't find subckt: " + modelname + '\n');
      return NULL;
    }else if(!dynamic_cast<const MODEL_SUBCKT*>(model)) {
      untested();
      error(bDANGER, modelname + " is not a subckt\n");
      return NULL;
    }
  }
  return model;
}
/*--------------------------------------------------------------------------*/
void DEV_SUBCKT::expand()
{
  const COMMON_SUBCKT* c = prechecked_cast<const COMMON_SUBCKT*>(common());
  assert(c);

  const CARD* model = find_subckt_model(this /*scope*/, c->modelname());
  if (!model) {
    error(bERROR, "no subckt\n");
  }
  new_subckt(model);
  subckt()->expand();

  assert(!is_constant()); /* because I have more work to do */
}
/*--------------------------------------------------------------------------*/
double DEV_SUBCKT::tr_probe_num(CS& cmd)const
{  
  {if (cmd.pmatch("V")) {
    untested();
    int nn = cmd.ctoi();
    untested();
    return (nn > 0 && nn <= net_nodes()) ? _n[nn-1].v0() : NOT_VALID;
  }else if (cmd.pmatch("P")) {
    untested();
    double power = 0.;
    untested();
    assert(subckt());
    untested();
    for (CARD_LIST::const_iterator
	   ci = subckt()->begin(); ci != subckt()->end(); ++ci) {
    untested();
      power += CARD::probe(*ci,"P");
    untested();
    }      
    untested();
    return power;
  }else if (cmd.pmatch("PD")) {
    untested(); 
    double power = 0.;
    untested();
    assert(subckt());
    untested();
    for (CARD_LIST::const_iterator
	   ci = subckt()->begin(); ci != subckt()->end(); ++ci) {
    untested();
      power += CARD::probe(*ci,"P");
    untested();
    }      
    untested();
    return power;
  }else if (cmd.pmatch("PS")) {
    untested(); 
    double power = 0.;
    untested();
    assert(subckt());
    untested();
    for (CARD_LIST::const_iterator
	   ci = subckt()->begin(); ci != subckt()->end(); ++ci) {
    untested();
      power += CARD::probe(*ci,"P");
    untested();
    }      
    untested();
    return power;
  }else{
    untested();
    return COMPONENT::tr_probe_num(cmd);
  }}
  /*NOTREACHED*/
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
